#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.quantization import quantize_tensor
from intel_npu_acceleration_library.nn.autograd import AutogradMatMul
from intel_npu_acceleration_library.backend import run_matmul
from typing import Optional, Union
import torch
import uuid


class Linear(torch.nn.Module):
    """Torch Linear operation NPU backend."""

    def __init__(self, weight: torch.Tensor, bias: Optional[torch.Tensor] = None):
        """Initialize the Linear class.

        Args:
            weight (torch.Tensor): Linear operation weight
            bias (Optional[torch.Tensor], optional): Linear operation optional bias. Defaults to None.
        """
        super().__init__()

        self.weight = torch.nn.Parameter(weight)
        self.bias = torch.nn.Parameter(bias) if isinstance(bias, torch.Tensor) else None
        self.outC, self.inC = self.weight.shape
        self.op_id = str(uuid.uuid4())
        # assert self.weight.dtype == torch.float16
        self._mm = AutogradMatMul.apply

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Torch module forward method.

        Args:
            x (torch.Tensor): Input tensor

        Returns:
            torch.Tensor: result
        """
        if self.training:
            out = self._mm(x, self.weight, None)
        else:
            out = run_matmul(x, self.weight, None, self.op_id)

        if self.bias is None:
            return out
        return out + self.bias

    @staticmethod
    def fromTorch(
        layer: torch.nn.Linear, dtype: torch.dtype = torch.float16
    ) -> Union["Linear", "QuantizedLinear"]:
        """Generate a NPU Linear layer from a torch one.

        Args:
            layer (torch.nn.Linear): the original torch.nn.Linear model to run on the NPU
            dtype (torch.dtype): the desired datatype

        Returns:
            Union[Linear, QuantizedLinear]: A NPU linear layer
        """
        if any(dim > 2**17 for dim in layer.weight.shape):
            return layer
        return Linear.fromTensor(layer.weight, getattr(layer, "bias", None), dtype)

    @staticmethod
    def fromTensor(
        weight: torch.Tensor,
        bias: Optional[torch.Tensor],
        dtype: torch.dtype = torch.float16,
    ) -> Union["Linear", "QuantizedLinear"]:
        """Generate a NPU Linear layer from a torch one.

        Args:
            weight (torch.Tensor): the original weight tensor
            bias (Optional[torch.Tensor]): the original bias tensor
            dtype (torch.dtype): the desired datatype

        Raises:
            RuntimeError: Quantized Linear requires input_channel to be a multiple of 8

        Returns:
            Union[Linear, QuantizedLinear]: A NPU linear layer
        """
        if dtype.is_floating_point:
            if bias is None:
                return Linear(weight.to(dtype), None)
            return Linear(weight.to(dtype), bias.to(dtype))
        elif dtype == torch.int8:
            if weight.shape[-1] % 8 != 0:
                raise RuntimeError(
                    "Quantized Linear requires input_channel to be a multiple of 8"
                )
            weights_quant, scale = quantize_tensor(weight)
            return QuantizedLinear(weights_quant, scale, bias)
        else:
            raise RuntimeError(
                f"intel-npu-acceleration-library library do not support yet the requeste datatype: {dtype}"
            )


class QuantizedLinear(torch.nn.Module):
    """Torch Quantized Linear operation NPU backend."""

    def __init__(
        self,
        weight: torch.Tensor,
        scale: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
    ):
        """Initialize the QuantizedLinear class.

        Args:
            weight (torch.Tensor): Linear operation weight
            scale (torch.Tensor): Quantization scale
            bias (Optional[torch.Tensor], optional): Linear operation optional bias. Defaults to None.

        Raises:
            RuntimeError: Quantized weight must be in torch.int8 format
        """
        super().__init__()

        self.weight = weight
        if self.weight.dtype != torch.int8:
            raise RuntimeError(
                f"Quantized weight must be in torch.int8 dtype instead of {self.weight.dtype}"
            )
        self.scale = scale
        self.outC, self.inC = self.weight.shape
        self.bias = bias
        self.op_id = str(uuid.uuid4())
        self._mm = AutogradMatMul.apply

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Torch module forward method.

        Args:
            x (torch.Tensor): Input tensor

        Raises:
            RuntimeError: Training is not supported for QuantizedLinear layer. Use `.eval()` to do inference only

        Returns:
            torch.Tensor: result
        """
        if self.training:
            raise RuntimeError(
                "Training is not supported for QuantizedLinear layer. Use `.eval()` to do inference only"
            )
        out = run_matmul(x, self.weight, self.scale, self.op_id)

        if self.bias is None:
            return out
        return out + self.bias



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.backend.ops import get_supported_ops
import numpy as np
import warnings
import ctypes
import sys
import os

handler = ctypes.POINTER(ctypes.c_char)
c_fp16_array = np.ctypeslib.ndpointer(dtype=np.float16, ndim=2, flags="C_CONTIGUOUS")
c_fp32_array = np.ctypeslib.ndpointer(dtype=np.float32, ndim=2, flags="C_CONTIGUOUS")
c_i8_array = np.ctypeslib.ndpointer(dtype=np.int8, ndim=2, flags="C_CONTIGUOUS")
c_u8_array = np.ctypeslib.ndpointer(dtype=np.uint8, ndim=2, flags="C_CONTIGUOUS")
c_u32_array = np.ctypeslib.ndpointer(dtype=np.uint32, ndim=1, flags="C_CONTIGUOUS")


def load_library() -> ctypes.CDLL:
    """Load the Intel® NPU Acceleration Library runtime library.

    Raises:
        RuntimeError: an error is raised if the platform is not supported. Currently supported platforms are WIndows and Linux

    Returns:
        ctypes.CDLL: The loaded dynamic library
    """
    path = os.path.dirname(os.path.abspath(__file__))
    if "openvino" in sys.modules:
        warnings.warn(
            "OpenVINO library is already loaded. It might interfere with NPU acceleration library if it uses an old version.",
            stacklevel=2,
        )

    external_path = os.path.join(path, "..", "external")
    sys.path.insert(0, external_path)

    if sys.platform == "win32":
        dll_path = os.path.join(path, "..", "lib", "Release")
        os.environ["OPENVINO_LIB_PATHS"] = dll_path
        os.add_dll_directory(os.path.abspath(dll_path))
        # Load DLL into memory.
        lib = ctypes.WinDLL(
            os.path.join(dll_path, "intel_npu_acceleration_library.dll")
        )  # , winmode=0)
    elif sys.platform == "linux":
        dll_path = os.path.join(path, "..", "lib")
        sys.path.append(dll_path)
        # In Linux it is required to explicitly load openvino lib
        _ = ctypes.CDLL(os.path.join(dll_path, "libopenvino.so"))
        lib = ctypes.CDLL(
            os.path.join(dll_path, "libintel_npu_acceleration_library.so")
        )
    else:
        raise RuntimeError(
            f"Platform {sys.platform} is not supported for intel-npu-acceleration-library library"
        )

    return lib


def init_common(lib: ctypes.CDLL):
    """Initialize common runtime bindings.

    Args:
        lib (ctypes.CDLL): Intel® NPU Acceleration Library runtime library
    """
    lib.saveModel.argtypes = [handler, ctypes.c_char_p]
    lib.saveCompiledModel.argtypes = [handler, ctypes.c_char_p]

    # Run a linar layer
    lib.run.argtypes = [handler, c_fp16_array, c_fp16_array]
    lib.run.restype = ctypes.c_float

    # Common destructor
    lib.destroyNNFactory.argtypes = [handler]

    lib.isNPUAvailable.restype = ctypes.c_bool


def init_network_factory(lib: ctypes.CDLL):
    """Initialize Netowrk factory bindings.

    Args:
        lib (ctypes.CDLL): Intel® NPU Acceleration Library runtime library
    """
    lib.createNNFactory.argtypes = [
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_bool,
    ]
    lib.createNNFactory.restype = handler

    lib.setNNFactoryWeights.argtypes = [handler, handler]

    lib.parameter.argtypes = [handler, ctypes.c_int, c_u32_array, ctypes.c_char_p]
    lib.parameter.restype = handler

    lib.compile.argtypes = [handler, handler]
    lib.compile.restype = handler

    lib.linear.argtypes = [
        handler,
        handler,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_bool,
        ctypes.c_char_p,
        ctypes.c_char_p,
    ]
    lib.linear.restype = handler

    for op in get_supported_ops():
        fn = getattr(lib, op.name)
        fn.argtypes = [handler] * (op.inputs + 1)
        fn.restype = handler


def init_parameters(lib: ctypes.CDLL):
    """Initialize Netowrk factory parameters.

    Args:
        lib (ctypes.CDLL): Intel® NPU Acceleration Library runtime library
    """
    lib.createParameters.argtypes = []
    lib.createParameters.restype = handler

    lib.destroyParameters.argtypes = [handler]

    lib.addFloatParameter.argtypes = [handler, c_fp16_array, ctypes.c_int, ctypes.c_int]
    lib.addIntParameter.argtypes = [
        handler,
        c_i8_array,
        c_fp16_array,
        ctypes.c_int,
        ctypes.c_int,
    ]

    lib.addInt4Parameter.argtypes = [
        handler,
        c_u8_array,
        c_fp16_array,
        ctypes.c_int,
        ctypes.c_int,
    ]

    lib.addIntParameterConversion.argtypes = [
        handler,
        c_i8_array,
        c_fp32_array,
        ctypes.c_int,
        ctypes.c_int,
    ]


def initialize_bindings() -> ctypes.CDLL:
    """Load the Intel® NPU Acceleration Library runtime library, and initialize all c++ <-> python bindings.

    Returns:
        ctypes.CDLL: Initialize matmul bindings
    """
    lib = load_library()

    init_common(lib)
    init_network_factory(lib)
    init_parameters(lib)

    return lib


lib = initialize_bindings()



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.backend.factory import NNFactory
from typing import Optional


class MLP(NNFactory):
    """Linear class, computing a matrix matrix multiplication with weights prefetching."""

    def __init__(
        self,
        hidden_size: int,
        intermediate_size: int,
        batch: int,
        activation: str = "swiglu",
        bias: Optional[bool] = False,
        profile: bool = False,
        device: str = "NPU",
    ):
        """Initialize the Linear class.

        Args:
            hidden_size (int): hidden_size channels
            intermediate_size (int): intermediate_size
            batch (int): batch
            activation (str): activation function to use
            bias (Optional[bool], optional): Enable/Disable bias. Defaults to False.
            profile (bool): Enable/Disable profiling. Defaults to False.
            device (str): Target device, default to "NPU".
        """
        super().__init__(hidden_size, hidden_size, batch, profile, device)

        mm1 = self.linear(self.input, intermediate_size, hidden_size, bias=bias)

        if activation == "swiglu":
            mm2 = self.linear(self.input, intermediate_size, hidden_size, bias=bias)  # type: ignore[attr-defined]
            mm1 = self.eltwise_mul(self.swish(mm1), mm2)  # type: ignore[attr-defined]
        else:
            atc_fn = getattr(self, activation)
            mm1 = atc_fn(mm1)

        out = self.linear(mm1, hidden_size, intermediate_size, bias=bias)
        self.compile(out)



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from transformers.models.llama.modeling_llama import (
    apply_rotary_pos_emb,
    repeat_kv,
    LlamaConfig,
)
from transformers import AutoTokenizer
from intel_npu_acceleration_library.nn import Linear
from intel_npu_acceleration_library.backend import run_factory, MLP
from functools import partial
from typing import Optional, List, Generator
from transformers.cache_utils import Cache
import torch
import uuid


class PhiMLP(torch.nn.Module):
    """Phi-2 MLP operation NPU backend."""

    def __init__(
        self,
        parameters: List[torch.Tensor],
    ):
        """Initialize LLAMA MLP operation.

        Args:
            parameters (List[torch.Tensor]): model weights
        """
        super().__init__()
        self.op_parameters = parameters
        self.op_id = str(uuid.uuid4())
        intermediate_size, hidden_size = parameters[0].shape
        self.backend_cls = partial(
            MLP,
            hidden_size=hidden_size,
            intermediate_size=intermediate_size,
            activation="gelu",
            bias=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Torch module forward method.

        Args:
            x (torch.Tensor): Input tensor

        Returns:
            torch.Tensor: result
        """
        return run_factory(x, self.op_parameters, self.backend_cls, self.op_id)

    @staticmethod
    def fromTorch(
        layer: torch.nn.Module, dtype: torch.dtype = torch.float16
    ) -> "PhiMLP":
        """Generate a NPU PhiMLP layer from a transformer one.

        Args:
            layer (torch.nn.Linear): the original PhiMLP model to run on the NPU
            dtype (torch.dtype): the desired datatype

        Returns:
            PhiMLP: A NPU PhiMLP layer
        """
        new_layer = PhiMLP(
            parameters=[weight.to(dtype) for weight in layer.parameters()],
        )

        return new_layer


class FusedLlamaMLP(torch.nn.Module):
    """LLAMA MLP operation NPU backend."""

    def __init__(
        self,
        parameters: List[torch.Tensor],
    ):
        """Initialize LLAMA MLP operation.

        Args:
            parameters (List[torch.Tensor]): model weights
        """
        super().__init__()
        self.op_parameters = parameters
        self.op_id = str(uuid.uuid4())
        intermediate_size, hidden_size = parameters[0].shape
        self.backend_cls = partial(
            MLP, hidden_size=hidden_size, intermediate_size=intermediate_size
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Torch module forward method.

        Args:
            x (torch.Tensor): Input tensor

        Returns:
            torch.Tensor: result
        """
        return run_factory(x, self.op_parameters, self.backend_cls, self.op_id)

    @staticmethod
    def fromTorch(
        layer: torch.nn.Module, dtype: torch.dtype = torch.float16
    ) -> "FusedLlamaMLP":
        """Generate a NPU LlamaMLP layer from a transformer LlamaMLP one.

        Args:
            layer (torch.nn.Linear): the original LlamaMLP model to run on the NPU
            dtype (torch.dtype): the desired datatype

        Returns:
            FusedLlamaMLP: A NPU LlamaMLP layer
        """
        new_layer = FusedLlamaMLP(
            parameters=[weight.to(dtype) for weight in layer.parameters()],
        )

        return new_layer


class LlamaAttention(torch.nn.Module):
    """LlamaAttention operation NPU backend."""

    def __init__(
        self,
        config: LlamaConfig,
        q_weights: torch.Tensor,
        kv_weights: torch.Tensor,
        o_proj: torch.Tensor,
        rotary_emb: torch.nn.Module,
        dtype: torch.dtype = torch.float16,
        layer_idx: Optional[int] = None,
    ):
        """Initialize the LlamaAttention class.

        Args:
            config (LlamaConfig): LlamaAttention configuration
            q_weights (torch.Tensor): Weights for the query Linear layer
            kv_weights (torch.Tensor): Concatentation of the weights for the Key and Value Linear layer
            o_proj (torch.Tensor): Weights for the output projection Linear layer
            rotary_emb (torch.nn.Module): Rotary embedding module
            dtype (torch.dtype): the desired datatype
            layer_idx (Optional[int], optional): Layer index. Defaults to None.
        """
        super().__init__()
        self.config = config
        self.rotary_emb = rotary_emb
        self.kv_proj = Linear.fromTensor(kv_weights, None, dtype)
        self.q_proj = Linear.fromTensor(q_weights, None, dtype)
        self.o_proj = Linear.fromTensor(o_proj, None, dtype)

        self.hidden_size = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.head_dim = self.hidden_size // self.num_heads
        self.num_key_value_heads = config.num_key_value_heads
        self.num_key_value_groups = self.num_heads // self.num_key_value_heads
        self.is_causal = True
        self.layer_idx = layer_idx

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        past_key_value: Optional[Cache] = None,
        output_attentions: Optional[bool] = False,
        use_cache: Optional[bool] = False,
        cache_position: Optional[torch.LongTensor] = None,
    ):
        """Torch module forward method.

        Args:
            hidden_states (torch.Tensor): input to the layer of shape `(batch, seq_len, embed_dim)`
            attention_mask (Optional[torch.Tensor], optional): attention mask of shape `(batch_size, sequence_length)`. Defaults to None.
            position_ids (Optional[torch.Tensor], optional): position_ids of shape `(batch_size, sequence_length)`. Defaults to None.
            past_key_value (Optional[Cache], optional): Pre-computed hidden-states (key and values in the self-attention blocks). Defaults to None.
            output_attentions (Optional[bool], optional):  Whether or not to return the attentions tensors of all attention layers.. Defaults to False.
            use_cache (Optional[bool], optional): If set to `True`, `past_key_values` key value states are returned. Defaults to False.
            cache_position (Optional[torch.LongTensor], optional): Cache position useful for static cache applications . Defaults to None.

        Returns:
            _type_: result
        """
        bsz, q_len, _ = hidden_states.size()

        query_states = self.q_proj(hidden_states)
        kv_states = self.kv_proj(hidden_states)

        key_states = kv_states[..., : self.num_key_value_heads * self.head_dim]
        value_states = kv_states[..., self.num_key_value_heads * self.head_dim :]

        query_states = query_states.view(
            bsz, q_len, self.num_heads, self.head_dim
        ).transpose(1, 2)
        key_states = key_states.view(
            bsz, q_len, self.num_key_value_heads, self.head_dim
        ).transpose(1, 2)
        value_states = value_states.view(
            bsz, q_len, self.num_key_value_heads, self.head_dim
        ).transpose(1, 2)

        cos, sin = self.rotary_emb(value_states, position_ids)

        query_states, key_states = apply_rotary_pos_emb(
            query_states, key_states, cos, sin, position_ids
        )

        if past_key_value is not None:
            # sin and cos are specific to RoPE models; position_ids needed for the static cache
            cache_kwargs = {"sin": sin, "cos": cos, "cache_position": cache_position}
            key_states, value_states = past_key_value.update(
                key_states, value_states, self.layer_idx, cache_kwargs
            )

        key_states = repeat_kv(key_states, self.num_key_value_groups)
        value_states = repeat_kv(value_states, self.num_key_value_groups)

        causal_mask = attention_mask
        if causal_mask is not None:
            causal_mask = causal_mask[:, :, :, : key_states.shape[-2]]

        attn_output = torch.nn.functional.scaled_dot_product_attention(
            query_states,
            key_states,
            value_states,
            attn_mask=causal_mask,
            is_causal=self.is_causal and attention_mask is None and q_len > 1,
        )

        attn_output = attn_output.transpose(1, 2).contiguous()
        attn_output = attn_output.reshape(bsz, q_len, self.hidden_size)

        attn_output = self.o_proj(attn_output)

        return attn_output, None, past_key_value

    @staticmethod
    def fromTorch(
        layer: torch.nn.Module, dtype: torch.dtype = torch.float16
    ) -> "LlamaAttention":
        """Generate a NPU LlamaAttention layer from a transformer LlamaAttention one.

        Args:
            layer (torch.nn.Linear): the original LlamaAttention model to run on the NPU
            dtype (torch.dtype): the desired datatype

        Returns:
            LlamaAttention: A NPU LlamaAttention layer
        """
        kv_weights = torch.cat((layer.k_proj.weight, layer.v_proj.weight), dim=0)

        new_layer = LlamaAttention(
            config=layer.config,
            q_weights=layer.q_proj.weight,
            kv_weights=kv_weights,
            o_proj=layer.o_proj.weight,
            rotary_emb=layer.rotary_emb,
            dtype=dtype,
            layer_idx=layer.layer_idx,
        )

        return new_layer


def lshift_insert(tensor: torch.Tensor, value: float) -> torch.Tensor:
    """Compute shift left and insert a value into a tensor.

    Args:
        tensor (torch.Tensor): input tensor
        value (float): value to add

    Returns:
        torch.Tensor: output tensor
    """
    tensor = torch.roll(tensor, shifts=-1, dims=-1)
    tensor[0, -1] = value
    return tensor


# Generate function
@torch.no_grad()
def generate_with_static_shape(
    model: torch.nn.Module,
    input_ids: torch.Tensor,
    max_length: int,
    attention_mask: Optional[torch.Tensor] = None,
    use_past: Optional[bool] = True,
    pad_token_id: Optional[int] = None,
    **kwargs,
) -> Generator[int, None, None]:
    """Run LLM generator routine wiht static shapes.

    Args:
        model (torch.nn.Module): LLM mode
        input_ids (torch.Tensor): model input_ids
        max_length (int): model max lenght.
        attention_mask (Optional[torch.Tensor], optional): input attention mask. Defaults to None.
        use_past (Optional[bool], optional): Enable/disable KV caching. Defaults to True.
        pad_token_id (Optional[int], optional): Padding token. Defaults to None.
        kwargs: Additional arguments

    Raises:
        RuntimeError: pad_token_id is not set and needed for static shape generation

    Yields:
        Generator[int, None, None]: Return a generator of new tokens
    """
    # Get sequence lenght
    batch, seq_lenght = input_ids.shape

    if pad_token_id is None:
        raise RuntimeError(
            "pad_token_id is not set and needed for static shape generation"
        )

    # Padding attention mask
    if attention_mask is None:
        attention_mask = torch.ones_like(input_ids, dtype=torch.int32).to(model.device)
    attention_mask_padding = torch.zeros(
        (batch, max_length - seq_lenght), dtype=input_ids.dtype, device=input_ids.device
    )
    attention_mask = torch.cat((attention_mask_padding, attention_mask), dim=-1)

    # Padding input_ids with left padding
    padding_input_ids = pad_token_id * torch.ones(
        (batch, max_length - seq_lenght), dtype=input_ids.dtype, device=input_ids.device
    )
    input_ids = torch.cat((padding_input_ids, input_ids), dim=-1).to(model.device)

    # Set the proper position ids
    position_ids = kwargs.get("position_ids", None)
    if position_ids is None:
        position_ids = torch.tensor(
            [[0] * (max_length - seq_lenght) + list(range(seq_lenght))],
            dtype=torch.int32,
        ).to(model.device)
    else:
        raise RuntimeError("Cannot set position_ids with in static shape generation")

    # past_key_values for KV-cache
    past_key_values = None

    for idx in range(seq_lenght, max_length):

        # Run the inference
        # position_ids=position_ids,
        out = model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
        )

        # Here I do greedy search as an example, but in general is where you want to select the next token with your fancy decoding algorithm
        logits = out.logits
        new_token = torch.argmax(logits[0, -1, :]).item()

        yield int(new_token)

        if not use_past:
            # Shift left input and position ids and set the new token and idx to the proper values
            input_ids = lshift_insert(input_ids, new_token)
            position_ids = lshift_insert(position_ids, idx)
        else:
            # Set input_ids and position_ids to their new value
            input_ids = torch.tensor([[new_token]], dtype=input_ids.dtype).to(
                model.device
            )
            position_ids = torch.tensor([[idx]], dtype=input_ids.dtype).to(model.device)

            # Select the proper KV cached keys for next inference
            past_key_values = [
                [item[:, :, 1:, :] for item in layer_past]
                for layer_past in out.past_key_values
            ]

        # Shift left attention mask and set the last value to one
        attention_mask = lshift_insert(attention_mask, 1)


def warm_up_decoder_model(
    tokenizer: AutoTokenizer,
    model: torch.nn.Module,
    model_seq_length: int,
    use_past: Optional[bool] = True,
):
    """Warm up the model on the NPU.

    This function JIT compile all the layers offloaded to the NPU and load and warm them into the NPU. This is particolarly useful for LLM decoders

    Args:
        tokenizer (AutoTokenizer): a tokenizer
        model (torch.nn.Module): a torch Module representing a language model decoder
        model_seq_length (int): Max sequence lenght for the tokenizer padding
        use_past (Optional[bool], optional): Enable or Disable KV-caching. Defaults to True.
    """
    input_ids = tokenizer(tokenizer.eos_token, return_tensors="pt")["input_ids"].to(
        "cpu"
    )

    results = generate_with_static_shape(
        model,
        input_ids=input_ids,
        max_length=model_seq_length,
        use_past=use_past,
        pad_token_id=tokenizer.pad_token_id,
    )
    idx = 0
    # Only two inferences
    for _ in results:
        if idx < 1:
            idx += 1
        else:
            break



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.optimizations import horizontal_fusion_linear
from torch._dynamo import register_backend
from typing import Union, Callable, Any

try:
    from transformers.models.llama.modeling_llama import LlamaMLP, LlamaAttention
    from transformers.models.gemma.modeling_gemma import GemmaMLP, GemmaAttention

    is_transformers_available = True
except ModuleNotFoundError:
    # Transformer library is not installed
    is_transformers_available = False


import intel_npu_acceleration_library.nn as nn
from typing import List
import torch


def compile(
    model: torch.nn.Module, dtype: torch.dtype = torch.float16, training: bool = False
) -> torch.nn.Module:
    """Compile a model for the NPU.

    Args:
        model (torch.nn.Module): a pytorch nn.Module to compile and optimize for the npu
        dtype (torch.dtype): the model target datatype, default to torch.float16
        training (bool): enable training. Default disabled

    Raises:
        RuntimeError: invalid datatypes

    Returns:
        torch.nn.Module: compiled NPU nn.Module
    """
    if not (dtype.is_floating_point or dtype == torch.int8):
        raise RuntimeError(
            f"intel-npu-acceleration-library library do not support yet the requeste datatype: {dtype}"
        )

    # Prepare and optimize model for NPU
    with torch.no_grad():
        # General optimizations
        apply_horizontal_fusion(model)
        optimize_llama_attention(model, dtype)

        # Model lowering to NPU ops
        lower_linear(model, dtype)

    if dtype.is_floating_point and training:
        # Set model to evaluation only as quantized training is not supported yet
        return model

    return model.eval()


def module_optimization(func: Callable) -> torch.nn.Module:
    """Optimize recursively a torch.nn.Module with a specific function.

    The function `func` get called recursively to every module in the network.

    Args:
        func (Callable): optimization function

    Returns:
        torch.nn.Module: optimized module
    """

    def wrapper(model: torch.nn.Module, *args: Any, **kwargs: Any):
        """Recursively apply the optimization function.

        Args:
            model (torch.nn.Module): original module
            args (Any): positional arguments
            kwargs (Any): keyword arguments

        """
        for name, layer in model.named_children():
            new_layer = func(name, layer, *args, **kwargs)
            if new_layer:
                model.add_module(name, new_layer)
                wrapper(new_layer, *args, **kwargs)
            else:
                wrapper(layer, *args, **kwargs)

    return wrapper


@module_optimization
def lower_linear(
    name: str, layer: torch.nn.Module, dtype: torch.dtype
) -> Union[torch.nn.Module, None]:
    """Lower torch.nn.Linear layer to NPU equivalent operators.

    Args:
        name (str): Layer name
        layer (torch.nn.Module): Original torch.nn.Linear module
        dtype (torch.dtype): Target datatype

    Returns:
        Union[torch.nn.Module, None]: Return the new NPU operator or None
    """
    if isinstance(layer, torch.nn.Linear):
        return nn.Linear.fromTorch(layer, dtype)
    if isinstance(layer, torch.nn.Conv2d):
        return nn.Conv2d.fromTorch(layer, dtype)
    return None


@module_optimization
def apply_horizontal_fusion(
    name: str, layer: torch.nn.Module
) -> Union[torch.nn.Module, None]:
    """Apply horizontal fusion (merging two linear layers with same input) when necessary.

    Args:
        name (str): Layer name
        layer (torch.nn.Module): Original module

    Returns:
        Union[torch.nn.Module, None]: optimized module
    """
    if isinstance(layer, (LlamaMLP, GemmaMLP)):
        return horizontal_fusion_linear(layer)
    return None


@module_optimization
def optimize_llama_attention(
    name: str, layer: torch.nn.Module, dtype: torch.dtype
) -> Union[torch.nn.Module, None]:
    """Optimize LLAMA attention block.

    Args:
        name (str): Module name
        layer (torch.nn.Module): Original Module
        dtype (torch.dtype): Target datatype

    Returns:
        Union[torch.nn.Module, None]: optimized llama module
    """
    if isinstance(layer, (LlamaAttention, GemmaAttention)):
        return nn.LlamaAttention.fromTorch(layer, dtype)
    return None


@register_backend
def npu(
    gm: Union[torch.nn.Module, torch.fx.GraphModule], example_inputs: List[torch.Tensor]
) -> Union[torch.nn.Module, torch.fx.GraphModule]:
    """Implement the custom torch 2.0 compile backend for the NPU.

    Args:
        gm (Union[torch.nn.Module, torch.fx.GraphModule]): The torch fx Module
        example_inputs (List[torch.Tensor]): A list of example inputs

    Returns:
        Union[torch.nn.Module, torch.fx.GraphModule]: The compiled model
    """
    # Run some optimizations
    gm = horizontal_fusion_linear(gm)

    # For now compile in fp16
    return compile(gm)



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from .bindings import lib
import subprocess  # nosec B404
import sys
import re


def npu_available() -> bool:
    """Return if the NPU is available.

    Returns:
        bool: Return True if the NPU is available in the system
    """
    return lib.isNPUAvailable()


def get_driver_version() -> str:
    """Get the driver version for the Intel® NPU Acceleration Library.

    Raises:
        RuntimeError: an error is raised if the platform is not supported. Currently supported platforms are Windows and Linux

    Returns:
        str: _description_
    """
    if not npu_available():
        raise RuntimeError("NPU is not available on this system")

    if sys.platform == "win32":
        out = (
            subprocess.check_output(  # nosec
                'powershell -Command " Get-WmiObject Win32_PnPSignedDriver | select devicename, driverversion',
                shell=True,
            )
            .decode("utf-8")
            .strip()
        )

        npu_drivers = [
            driver for driver in out.split("\n") if "Intel(R) AI Boost" in driver
        ]
        if len(npu_drivers) == 0:
            raise RuntimeError("Cannot get driver version")
        elif len(npu_drivers) > 1:
            raise RuntimeError("Multiple drivers found")
        else:
            driver_info = re.search(r"(\d+\.\d+\.\d+\.\d+)", npu_drivers[0])
            if driver_info:
                driver_version = driver_info.group(1)
                return driver_version
            else:
                raise RuntimeError("Cannot get driver version")
    else:
        raise RuntimeError(f"Cannot get driver version for {sys.platform}")



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from .linear import Linear, QuantizedLinear  # noqa
from .conv import Conv2d  # noqa

try:
    from .llm import LlamaAttention, PhiMLP  # noqa

    llm_modules = ["LlamaAttention", "PhiMLP"]
except ModuleNotFoundError:
    # Transformer library is not installed
    llm_modules = []


__all__ = ["Linear", "QuantizedLinear", "Conv2d"] + llm_modules



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#
from typing import Tuple
import torch


def quantize_tensor(
    weight: torch.Tensor, min_max_range: Tuple[int, int] = (-128, 127)
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Quantize a fp16 tensor symmetrically.

    Produces a quantize tensor (same shape, dtype == `torch.int8`) and a scale tensor (dtype == `torch.float16)
    The quantization equation is the following W = S * W_q

    Args:
        weight (torch.Tensor): The tensor to quantize
        min_max_range (Tuple[int, int]): The min and max range for the quantized tensor. Defaults to (-128, 127).

    Raises:
        RuntimeError: Error in the quantization step

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Quantized tensor and scale
    """
    scale = torch.max(torch.abs(weight), dim=-1).values

    # if any of the elements are zeros set the scale to the max value
    if torch.any(scale == 0):
        scale = torch.ones_like(scale) * torch.max(torch.abs(weight))

    # Compute scale and zero point
    scale = (scale / max(min_max_range)).to(torch.float16).view(-1, 1)

    weights_quant = torch.floor(weight / scale)

    if not (
        torch.max(weights_quant) <= max(min_max_range)
        and torch.min(weights_quant) >= min(min_max_range)
    ):
        raise RuntimeError(
            f"Quantization error: range of quantized weghts = {(torch.min(weights_quant), torch.max(weights_quant))} instead of ({min_max_range})"
        )
    return weights_quant.to(torch.int8), scale


def compress_to_i4(weights: torch.Tensor) -> torch.Tensor:
    """
    Compresses a given tensor to 4-bit representation.

    Args:
        weights (torch.Tensor): The input tensor to be compressed.

    Returns:
        torch.Tensor: The compressed tensor with 4-bit representation.
    """
    compressed_weights = torch.zeros(
        (weights.shape[0], weights.shape[1] // 2), dtype=torch.uint8
    )
    for i in range(weights.shape[1] // 2):
        compressed_weights[:, i] = (weights[:, 2 * i] & 0x0F) | (
            ((weights[:, 2 * i + 1] & 0x0F) << 4) & 0xF0
        )
    return compressed_weights



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#
from .bindings import lib
from .utils import npu_available, get_driver_version
from .mlp import MLP
from .matmul import MatMul
from .linear import Linear
from .qmatmul import QMatMul
from .qlinear import QLinear
from .factory import NNFactory
from .runtime import run_matmul, run_factory, clear_cache
import warnings


if not npu_available():
    warnings.warn(
        "NPU is not available in your system. Library will fallback to AUTO device selection mode",
        stacklevel=2,
    )


__all__ = [
    "NNFactory",
    "MLP",
    "MatMul",
    "Linear",
    "QMatMul",
    "QLinear",
    "run_matmul",
    "run_factory",
    "clear_cache",
    "npu_available",
    "get_driver_version",
    "lib",
]



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#
from typing import Optional, List, Union, Any, Dict, Tuple, Iterable
from intel_npu_acceleration_library.backend.bindings import lib as backend_lib
import numpy as np
import intel_npu_acceleration_library
import ctypes
import os


def adapt_weight(w: np.ndarray) -> np.ndarray:
    """Adapt the weights to run on the NPU.

    Args:
        w (np.ndarray): weights array

    Raises:
        RuntimeError: Unsupported shape

    Returns:
        np.ndarray: The adapted array
    """
    if len(w.shape) == 1:
        return w.reshape((1, -1))
    elif len(w.shape) == 2:
        return w
    else:
        raise RuntimeError(f"Unnupported shape > 2: {w.shape}")


class BaseNPUBackend:
    """A base class that represent a abstract Matrix-Matrix operation on the NPU."""

    def __init__(self, profile: Optional[bool] = False) -> None:
        """Initialize class profiling.

        Args:
            profile (Optional[bool], optional): Enable/Disable NPU profiling. Defaults to False.
        """
        if profile:
            os.environ["NPU_PRINT_PROFILING"] = "JSON"
            os.environ["NPU_PROFILING_OUTPUT_FILE"] = "profiling.json"
            os.environ["NPU_PROFILING_VERBOSITY"] = "HIGH"
        self._mm: Any = None

    def __del__(self):
        """Deallocate and free the class from the library."""
        if (
            hasattr(self, "_mm")
            and intel_npu_acceleration_library
            and hasattr(backend_lib, "destroyNNFactory")
        ):
            backend_lib.destroyNNFactory(self._mm)

    def save(self, path: str):
        """Save the Openvino model.

        Args:
            path (str): the model save path
        """
        backend_lib.saveModel(self._mm, ctypes.c_char_p(path.encode()))

    def saveCompiledModel(self, path: str):
        """Save the compiled model.

        Args:
            path (str): the compiled model save path
        """
        backend_lib.saveCompiledModel(self._mm, ctypes.c_char_p(path.encode()))


class BaseNPUBackendWithPrefetch(BaseNPUBackend):
    """A base class that represent a abstract Matrix-Matrix operation on the NPU.

    Linear type classes employ an algorithm to optimize weights prefetching
    """

    def __init__(self, profile: bool):
        """Initialize class.

        Args:
            profile (bool): Enable/Disable NPU profiling.
        """
        super().__init__(profile)
        self.wt_order: List[str] = []
        self.wt_map: Dict[str, ctypes._Pointer] = {}
        self.loaded: Optional[str] = None

    def load_wt_fn(self, module, parameters):
        """Load asyncronously the parameter into the NPU.

        Args:
            module: the NPU backend module
            parameters: the weights parameter class
        """
        backend_lib.setNNFactoryWeights(module, parameters)

    def create_parameters(
        self, weights: Iterable[Union[np.ndarray, Tuple[np.ndarray, ...]]]
    ) -> ctypes._Pointer:
        """Create an operation parameter from a list of weights.

        Args:
            weights (Iterable[Union[np.ndarray, Tuple[np.ndarray, ...]]]): Operation parameters

        Raises:
            RuntimeError: Quantized weights needs to be in int8 format
            ValueError: Invalid dtype for scale

        Returns:
            ctypes._Pointer: an instance to the Parameters object
        """
        param = backend_lib.createParameters()
        if isinstance(weights, (list, tuple)):
            for weight in weights:
                if isinstance(weight, (list, tuple)):
                    # int8: data and scale
                    data, scale = weight
                    if data.dtype not in [np.int8, np.uint8]:
                        raise RuntimeError(
                            "Quantized weights needs to be in int8 or uint8 format"
                        )
                    shape = data.shape if len(data.shape) > 1 else [data.shape[0], 1]
                    if scale.dtype == np.float16:
                        # Mixed precision matmul
                        if data.dtype == np.int8:
                            backend_lib.addIntParameter(
                                param,
                                adapt_weight(data),
                                adapt_weight(scale),
                                *shape,
                            )
                        else:
                            backend_lib.addInt4Parameter(
                                param,
                                adapt_weight(data),
                                adapt_weight(scale),
                                *shape,
                            )
                    elif scale.dtype == np.float32:
                        # FP16 matmul with CPU conversion
                        backend_lib.addIntParameterConversion(
                            param,
                            adapt_weight(data),
                            adapt_weight(scale),
                            *shape,
                        )
                    else:
                        raise ValueError(f"Invalid dtype for scale: {scale.dtype}")
                else:
                    shape = (
                        weight.shape if len(weight.shape) > 1 else [weight.shape[0], 1]
                    )
                    backend_lib.addFloatParameter(param, adapt_weight(weight), *shape)
        elif isinstance(weights, np.ndarray):
            shape = weights.shape if len(weights.shape) > 1 else [weights.shape[0], 1]
            backend_lib.addFloatParameter(param, adapt_weight(weights), *shape)
        return param

    def add_to_map(
        self, wt_hash: str, weights: Iterable[Union[np.ndarray, Tuple[np.ndarray, ...]]]
    ):
        """Add an operation parameters to the operation hash:parameter map.

        Args:
            wt_hash (str): operation hash
            weights (Iterable[Union[np.ndarray, Tuple[np.ndarray, ...]]]): Operation parameters
        """
        self.wt_map[wt_hash] = self.create_parameters(weights)

        self.wt_order.append(wt_hash)

    def setWeights(
        self, wt_hash: Optional[str], *args: Union[np.ndarray, Tuple[np.ndarray, ...]]
    ) -> bool:
        """Set the operation weights in the NPU.

        Args:
            wt_hash (str): operation hash. If set to None force the load of the weights
            args (Union[np.ndarray, Tuple[np.ndarray, ...]]): Variable length weights list. Can be a np array or a tuple of weight, scale in case of quantized tensors

        Returns:
            bool: Return True if the op parameters are already in the op map
        """
        if wt_hash is None:
            self.load_wt_fn(self._mm, self.create_parameters(args))
            return False
        in_wt_map = wt_hash in self.wt_map.keys()
        if not wt_hash == self.loaded:
            if not in_wt_map:
                self.add_to_map(wt_hash, args)
            self.load_wt_fn(self._mm, self.wt_map[wt_hash])
            self.loaded = wt_hash
            return in_wt_map
        return in_wt_map

    def prefetchWeights(self):
        """Prefetch next operation weights."""
        next_wt_idx = (self.wt_order.index(self.loaded) + 1) % len(self.wt_order)
        wt_hash = self.wt_order[next_wt_idx]
        if not wt_hash == self.loaded:
            self.load_wt_fn(self._mm, self.wt_map[wt_hash])
            self.loaded = wt_hash

    def __del__(self):
        """Deallocate and free the class from the library."""
        super(BaseNPUBackendWithPrefetch, self).__del__()
        for par in self.wt_map.values():
            if intel_npu_acceleration_library and hasattr(
                backend_lib, "destroyParameters"
            ):
                backend_lib.destroyParameters(par)



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from dataclasses import dataclass
from functools import lru_cache
from typing import List


@dataclass(frozen=True)
class SupportedOp:
    """A class for supported runtime OPs in the NPU.

    Attrs:
        name (str): Operation name
        inputs (int): Number of inputs
    """

    name: str
    inputs: int


@lru_cache(maxsize=None)
def get_supported_ops() -> List[SupportedOp]:
    """Generate a list fo supported operations.

    Returns:
        List[SupportedOp]: list fo supported NPU operations
    """
    supported_ops = [
        SupportedOp(name="matmul", inputs=2),
        SupportedOp(name="eltwise_add", inputs=2),
        SupportedOp(name="eltwise_mul", inputs=2),
        SupportedOp(name="eltwise_div", inputs=2),
        SupportedOp(name="gelu", inputs=1),
        SupportedOp(name="softmax", inputs=1),
        SupportedOp(name="swish", inputs=1),
        SupportedOp(name="convert_to_fp16", inputs=1),
    ]
    return supported_ops



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

import intel_npu_acceleration_library.nn as nn
import torch


class Conv2d(torch.nn.Module):
    """
    2D convolutional layer implementation.

    Attrs:
        weight (torch.Tensor): The weight tensor of the layer.
        bias (torch.Tensor): The bias tensor of the layer.

    Args:
        matmul (torch.nn.Module): The matrix multiplication module.
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (Union[int, Tuple[int, int]]): Size of the convolutional kernel.
        stride (Union[int, Tuple[int, int]], optional): Stride of the convolution. Defaults to (1, 1).
        padding (Union[int, Tuple[int, int]], optional): Padding added to the input. Defaults to (0, 0).
        dilation (Union[int, Tuple[int, int]], optional): Dilation rate of the convolution. Defaults to (1, 1).
    """

    def __init__(
        self,
        matmul,
        in_channels,
        out_channels,
        kernel_size,
        stride=(1, 1),
        padding=(0, 0),
        dilation=(1, 1),
    ) -> None:
        """Initialize a Convolutional layer.

        Args:
            matmul: The matrix multiplication function to be used.
            in_channels: The number of input channels.
            out_channels: The number of output channels.
            kernel_size: The size of the convolutional kernel.
            stride: The stride of the convolution. Defaults to (1, 1).
            padding: The padding added to the input. Defaults to (0, 0).
            dilation: The dilation rate of the convolution. Defaults to (1, 1).
        """
        super().__init__()

        self.matmul = matmul
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.dilation = dilation
        self.stride = stride

    @property
    def weight(self) -> torch.Tensor:
        """
        Get the weight tensor of the layer.

        Returns:
            torch.Tensor: The weight tensor.
        """
        return self.matmul.weight

    @property
    def bias(self) -> torch.Tensor:
        """
        Get the bias tensor of the layer.

        Returns:
            torch.Tensor: The bias tensor.
        """
        return self.matmul.bias

    def compute_output_dim(self, dim, idx) -> int:
        """
        Compute the output dimension for a given input dimension.

        Args:
            dim (int): Input dimension.
            idx (int): Index of the dimension.

        Returns:
            int: Output dimension.
        """
        return (
            dim
            + 2 * self.padding[idx]
            - self.dilation[idx] * (self.kernel_size[idx] - 1)
            - 1
        ) // self.stride[idx] + 1

    def forward(self, x) -> torch.Tensor:
        """
        Forward pass of the convolutional layer.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        # Unfold the input
        inp_unf = torch.nn.functional.unfold(
            x, self.kernel_size, self.dilation, self.padding, self.stride
        ).transpose(1, 2)
        out_unf = self.matmul(inp_unf).transpose(1, 2)

        out_shape = [x.shape[0], self.out_channels] + [
            self.compute_output_dim(dim, idx) for idx, dim in enumerate(x.shape[2:])
        ]
        out = out_unf.view(out_shape)

        return out

    @staticmethod
    def fromTorch(layer, dtype) -> "Conv2d":
        """
        Create a Conv2d layer from a torch.nn.Conv2d layer.

        Args:
            layer (torch.nn.Conv2d): The torch Conv2d layer.
            dtype (torch.dtype): Data type of the layer.

        Returns:
            Conv2d: The converted Conv2d layer.
        """
        weight = layer.weight.view(layer.weight.shape[0], -1)
        matmul = nn.Linear.fromTensor(weight, getattr(layer, "bias", None), dtype)
        new_layer = Conv2d(
            matmul,
            layer.in_channels,
            layer.out_channels,
            layer.kernel_size,
            layer.stride,
            layer.padding,
            layer.dilation,
        )

        return new_layer



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.backend import run_matmul
from typing import Optional, Iterable, Union
import torch


class AutogradMatMul(torch.autograd.Function):
    """Autograd module for Linear operation."""

    @staticmethod
    def forward(
        ctx, x: torch.Tensor, w: torch.Tensor, scale: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """Run a linear forward pass. Depending on the datatype of the weights it runs a float or quantized operation.

            Equivalent pytorch code:
            result = x @ w.T

        Args:
            ctx (Any): the autograd context
            x (torch.Tensor): Activation tensor. Its dtype must be torch.float16
            w (torch.Tensor): Weight tensor. Its dtype must be torch.float16
            scale (Optional[torch.Tensor], optional): Quantization scale. If weights.dtype == torch.int8 then it must be set. Defaults to None.

        Returns:
            torch.Tensor: result
        """
        result = run_matmul(x, w, scale, None)
        ctx.save_for_backward(w, x)
        return result

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> Iterable[Union[torch.Tensor, None]]:
        """Run a linear backward pass.

        grad_output shape: [batch, output_channels]
        x shape: [batch, input_channels]
        w shape: [output_channels, input_channels]

        Expected gradients
        dl_dx shape: [batch, input_channels]
        dl_dw shape: [output_channels, input_channels]

        Equivalent pytorch code:
        dl_dx = grad_output @ w.to(torch.float32)
        dl_dw =  (x.T @ grad_output).T

        Args:
            ctx (Any): the autograd context
            grad_output (torch.Tensor): output gradient

        Returns:
            Iterable[Union[torch.Tensor, None]]: Input and parameters gradients
        """
        (
            w,
            x,
        ) = ctx.saved_tensors

        dl_dx = run_matmul(grad_output, torch.transpose(w, -1, -2))
        dl_dw = run_matmul(
            torch.transpose(grad_output, -1, -2), torch.transpose(x, -1, -2)
        )
        return dl_dx, dl_dw, None



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.backend import Linear, QLinear
from intel_npu_acceleration_library.backend import MatMul, QMatMul
from intel_npu_acceleration_library.backend import NNFactory
from torch.profiler import record_function
from typing import Optional, List, Any, Dict, Deque
from collections import deque
import numpy as np
import torch

_model_cache: Dict[str, Deque[NNFactory]] = {}


def clear_cache():
    """Clear the cache of models."""
    global _model_cache
    _model_cache = {}


@torch.no_grad()
def run_matmul(
    x: torch.Tensor,
    weights: torch.Tensor,
    scale: Optional[torch.Tensor] = None,
    op_id: Optional[str] = None,
) -> torch.Tensor:
    """Run a matmul operation. Depending on the datatype of the weights it runs a float or quantized operation.

    Args:
        x (torch.Tensor): Activation tensor. Its dtype must be torch.float16
        weights (torch.Tensor): Weights tensor.  Its dtype can be torch.float16 or torch.int8
        scale (Optional[torch.Tensor], optional): Quantization scale. If weights.dtype == torch.int8 then it must be set. Defaults to None.
        op_id (Optional[str], optional): Operation ID. Defaults to None.

    Raises:
        RuntimeError: Unsupported weights datatype. Supported types: [torch.float16, torch.int8]

    Returns:
        torch.Tensor: result
    """
    global _model_cache

    outC, inC = weights.shape[-2:]

    # Set tensors as contiguous in memory
    x = set_contiguous(x)
    weights = set_contiguous(weights)
    weights = weights.view([-1, weights.shape[-1]])

    if weights.dtype.is_floating_point:
        op_class = Linear if op_id is not None else MatMul
        op_args = [weights.to(torch.float16).numpy()]
    elif weights.dtype == torch.int8:
        if scale is None:
            raise RuntimeError("Quantized weights require a not null scale")
        op_class = QLinear if op_id is not None else QMatMul
        if scale is None:
            raise RuntimeError(
                f"Quantized matmul (weights dtype == {weights.dtype}) requires scale (scale = {scale})"
            )
        op_args = [weights.numpy(), scale.numpy()]
    else:
        raise RuntimeError(f"Unsupported dtype for weights {weights.dtype}")

    if not x.dtype.is_floating_point:
        raise RuntimeError(f"Unsupported dtype for activation {x.dtype}")

    # Use or not op_id depending on the class used
    op_kwargs = {"op_id": op_id} if op_id else {}

    original_input_shape = x.shape
    expected_output_shape = list(original_input_shape[:-1]) + [outC]

    if not (len(x.shape) >= 2):
        raise RuntimeError(f"Input shape {x.shape} must me >= 2")

    # Reshape input
    input_dtype = x.dtype
    x_np = x.to(torch.float16).view([-1, inC]).numpy()

    real_batch = x_np.shape[0]

    # If the real batch is 1, we need to use 16 and then slice it
    if real_batch == 1:
        batch = 1
    else:
        batch = real_batch

    key = f"{str(op_class.__name__)}_{batch}_{inC}_x_{outC}_{inC}_{x_np.dtype}"
    models = _model_cache.get(key, None)

    if models is None:
        _model_cache[key] = deque([op_class(inC, outC, batch)])
    elif len(models) < 1:
        _model_cache[key].append(op_class(inC, outC, batch))
    else:
        _model_cache[key].rotate(1)

    # Get the model
    model = _model_cache[key][0]

    if real_batch == 1:
        # Expand and then slice
        with record_function(f"npu_matvec_{key}"):
            ret = model.run(np.vstack(1 * [x_np]), *op_args, **op_kwargs)[:1, ...]
    else:
        with record_function(f"npu_matmul_{key}"):
            ret = model.run(x_np, *op_args, **op_kwargs)
    return torch.from_numpy(ret).view(expected_output_shape).to(input_dtype)


def set_contiguous(tensor: torch.Tensor) -> torch.Tensor:
    """Set tensor to be contiguous in memory.

    Args:
        tensor (torch.Tensor): input tensor

    Returns:
        torch.Tensor: output, contiguous tensor
    """
    if not tensor.is_contiguous():
        return tensor.contiguous()
    return tensor


@torch.no_grad()
def run_factory(
    x: torch.Tensor,
    weights: List[torch.Tensor],
    backend_cls: Any,
    op_id: Optional[str] = None,
) -> torch.Tensor:
    """Run a factory operation. Depending on the datatype of the weights it runs a float or quantized operation.

    Args:
        x (torch.Tensor): Activation tensor. Its dtype must be torch.float16
        weights (torch.Tensor): Weights tensor.  Its dtype can be torch.float16 or torch.int8
        backend_cls (Any): Backend class to run
        op_id (Optional[str], optional): Operation ID. Defaults to None.

    Returns:
        torch.Tensor: result
    """
    global _model_cache

    inC = x.shape[-1]
    # TODO: fix this
    outC = inC

    # Use or not op_id depending on the class used
    op_kwargs = {"op_id": op_id} if op_id else {}

    original_input_shape = x.shape
    expected_output_shape = list(original_input_shape[:-1]) + [outC]

    # Reshape input
    input_dtype = x.dtype
    x_np = x.to(torch.float16).view((-1, inC)).numpy()

    op_args = [w.to(torch.float16).numpy() for w in weights]

    real_batch = x_np.shape[0]

    # If the real batch is 1, we need to use 16 and then slice it
    if real_batch == 1:
        batch = 16
    else:
        batch = real_batch

    key = f"{backend_cls.func.__name__}_{batch}_{inC}_{outC}_{x_np.dtype}"
    models = _model_cache.get(key, None)

    if models is None:
        _model_cache[key] = deque([backend_cls(batch=batch)])
    elif len(models) < 1:
        _model_cache[key].append(backend_cls(batch=batch))
    else:
        _model_cache[key].rotate(1)

    # Get the model
    model = _model_cache[key][0]

    if real_batch == 1:
        # Expand and then slice
        with record_function(f"npu_factory_vect_{key}"):
            ret = model.run(np.vstack(16 * [x_np]), *op_args, **op_kwargs)[:1, ...]
    else:
        with record_function(f"npu_factory_mul_{key}"):
            ret = model.run(x_np, *op_args, **op_kwargs)

    return torch.from_numpy(ret).view(expected_output_shape).to(input_dtype)



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.backend.base import BaseNPUBackendWithPrefetch
from intel_npu_acceleration_library.backend.ops import get_supported_ops
from intel_npu_acceleration_library.backend.bindings import lib as backend_lib
from typing import Optional, Tuple, Any, Union
from functools import partial
import numpy.typing as npt
import numpy as np
import ctypes


class NNFactory(BaseNPUBackendWithPrefetch):
    """Linear class, computing a matrix matrix multiplication with weights prefetching."""

    def __init__(
        self,
        inC: int,
        outC: int,
        batch: int,
        profile: bool = False,
        device: str = "NPU",
    ):
        """Initialize the Linear class.

        Args:
            inC (int): input channels
            outC (int): output channels
            batch (int): batch
            profile (Optional[bool], optional): Enable/Disable profiling. Defaults to False.
            device (str): Target device, default to "NPU".
        """
        super().__init__(profile)
        self.batch = batch
        self.inC, self.outC = inC, outC
        self.device = device
        self._mm = backend_lib.createNNFactory(
            ctypes.c_char_p(self.device.encode()),
            self.inC,
            self.outC,
            self.batch,
            profile,
        )
        self.out = np.empty((self.batch, self.outC), dtype=np.float16)
        self.elapsed = None
        self.input = self.parameter((self.batch, self.inC))

        for op in get_supported_ops():
            setattr(self, op.name, partial(self._call_backend_op, op.name))

    def _call_backend_op(self, op_name: str, *parameters: Any) -> Any:
        """Dynamically call a backend operation.

        Args:
            op_name (str): operation name
            parameters (Any): variable list of operation parameters

        Returns:
            Any: Operation
        """
        fn = getattr(backend_lib, op_name)
        return fn(self._mm, *parameters)

    def get_backend_dtype(self, dtype) -> ctypes.c_char_p:
        """Get the string representation of the dtype.

        Args:
            dtype: numpy dtype

        Raises:
            RuntimeError: Unsupported datatype

        Returns:
            ctypes.c_char_p: string representation of the dtype
        """
        if dtype == np.int8:
            str_dtype = "int8"
        elif dtype == np.uint8:
            # u8 represents packed i4 dtypes
            str_dtype = "int4"
        elif dtype == np.float16:
            str_dtype = "float16"
        else:
            raise RuntimeError(f"DType is not supported {dtype}")
        return ctypes.c_char_p(str_dtype.encode())

    def parameter(
        self, shape: Tuple[int, int], dtype: npt.DTypeLike = np.float16
    ) -> ctypes._Pointer:
        """Generate a model input parameter.

        Args:
            shape (Tuple[int, int]): Parameter shape (only 2D tensors supported atm)
            dtype (np.dtype, optional): parameter type np.int8, np.uint8 and np.float16 supported. Defaults to np.float16. Unit8 represents packed i4 dtypes

        Returns:
            ctypes._Pointer: an instance to a parameter object

        """
        shape_ptr = np.array(shape, dtype=np.uint32)
        return backend_lib.parameter(
            self._mm, shape_ptr.size, shape_ptr, self.get_backend_dtype(dtype)
        )

    def linear(
        self,
        input_node: ctypes._Pointer,
        output_channels: int,
        input_channels: int,
        bias: Optional[bool] = False,
        act_dtype: npt.DTypeLike = np.float16,
        wt_dtype: npt.DTypeLike = np.float16,
    ) -> ctypes._Pointer:
        """Generate a linear layer.

        Args:
            input_node (ctypes._Pointer): layer input node
            output_channels (int): number of output channels
            input_channels (int): number of input channels
            bias (bool, optional): enable/disable bias. Defaults to False.
            act_dtype (npt.DTypeLike, optional): activation dtype. Defaults to np.float16.
            wt_dtype (npt.DTypeLike, optional): weight dtype. Defaults to np.float16.

        Returns:
            ctypes._Pointer: _description_
        """
        return backend_lib.linear(
            self._mm,
            input_node,
            output_channels,
            input_channels,
            bias,
            self.get_backend_dtype(act_dtype),
            self.get_backend_dtype(wt_dtype),
        )

    def compile(self, output_node: ctypes._Pointer):
        """Finalize and compile a model.

        Args:
            output_node (ctypes._Pointer): Model output node
        """
        backend_lib.compile(self._mm, output_node)

    def run(
        self,
        X: np.ndarray,
        *weights: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
        **kwargs: Any,
    ) -> np.ndarray:
        """Run the layer: X * W^T.

        Args:
            X (np.ndarray): lhs operator
            weights (Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]): rhs operators
            kwargs (Any): additional arguments

        Raises:
            RuntimeError: Input tensor shape mismatch

        Returns:
            np.ndarray: result
        """
        if not (X.shape[0] == self.batch and X.shape[1] == self.inC):
            raise RuntimeError(
                f"Input shape {X.shape} different from expected one {(self.batch, self.inC)}"
            )

        prefetch = self.setWeights(kwargs.get("op_id", None), *weights)

        self.elapsed = backend_lib.run(self._mm, X, self.out)

        if prefetch:
            self.prefetchWeights()

        return self.out



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig
import pathlib
import glob
import os


class CMakeExtension(Extension):
    def __init__(self, name):
        # don't invoke the original build_ext for this special extension
        headers = glob.glob("include/**/*.h")
        cpp_sources = glob.glob("src/*.cpp")
        requirements = glob.glob("*requirements.txt")
        sources = ["CMakeLists.txt"] + requirements + cpp_sources + headers
        super().__init__(name, sources=sources)


class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)
        super().run()

    def build_cmake(self, ext):
        cwd = pathlib.Path().absolute()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        extdir.mkdir(parents=True, exist_ok=True)

        # example of cmake args
        config = "Debug" if self.debug else "Release"
        cmake_args = [
            f'-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={os.path.join(extdir.parent.absolute(), ext.name, "lib")}',
            "-DCMAKE_BUILD_TYPE=" + config,
        ]

        # example of build args
        build_args = [
            "--config",
            config,
        ]

        os.chdir(str(build_temp))
        self.spawn(["cmake", str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(["cmake", "--build", "."] + build_args)
        # Troubleshooting: if fail on line above then delete all possible
        # temporary CMake files including "CMakeCache.txt" in top level dir.
        os.chdir(str(cwd))


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    requirements = fh.readlines()

with open("dev_requirements.txt") as fh:
    dev_requirements = fh.readlines()

setup(
    name="intel_npu_acceleration_library",
    version="v1.1.0",
    packages=[
        "intel_npu_acceleration_library",
        "intel_npu_acceleration_library.backend",
        "intel_npu_acceleration_library.nn",
    ],
    author="Alessandro Palla",
    author_email="alessandro.palla@intel.com",
    description="Intel® NPU Acceleration Library",
    license="Apache License 2.0",
    url="https://github.com/intel/intel-npu-acceleration-library",
    ext_modules=[CMakeExtension("intel_npu_acceleration_library")],
    cmdclass={
        "build_ext": build_ext,
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    install_requires=requirements,
    extras_require={
        "dev": dev_requirements,
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="intel-npu-acceleration-library, machine learning, llm, intel core ultra",
)



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#

from intel_npu_acceleration_library.backend.factory import NNFactory
import numpy as np


class Linear(NNFactory):
    """Linear class, computing a matrix matrix multiplication with weights prefetching."""

    def __init__(
        self,
        inC: int,
        outC: int,
        batch: int,
        profile: bool = False,
        device: str = "NPU",
    ):
        """Initialize the Linear class.

        Args:
            inC (int): input channels
            outC (int): output channels
            batch (int): batch
            profile (bool): Enable/Disable profiling. Defaults to False.
            device (str): Target device, default to "NPU".
        """
        super().__init__(inC, outC, batch, profile, device)
        out = self.linear(self.input, outC, inC, bias=False)
        self.compile(out)

    def run(self, X: np.ndarray, W: np.ndarray, op_id: str) -> np.ndarray:
        """Run the layer: X * W^T.

        Args:
            X (np.ndarray): lhs operator
            W (np.ndarray): rhs operator
            op_id (str): operation id

        Raises:
            RuntimeError: Input or weight tensor shape mismatch

        Returns:
            np.ndarray: result
        """
        if not (X.shape[0] == self.batch and X.shape[1] == self.inC):
            raise RuntimeError(
                f"Input shape {X.shape} different from expected one {(self.batch, self.inC)}"
            )
        if not (X.shape[0] == self.batch and X.shape[1] == self.inC):
            raise RuntimeError(
                f"Weight shape {W.shape} different from expected one {(self.outC, self.inC)}"
            )

        return super().run(X, W, op_id=op_id)



#
# Copyright © 2024 Intel Corporation
# SPDX-License-Identifier: Apache 2.0
#
from typing import Dict, List, Any
import torch.nn as nn
import torch.fx as fx
import operator
import torch


def delattr_recursively(module: nn.Module, target: str):
    """Delete attribute recursively by name in a torch.nn.Module.

    Args:
        module (nn.Module): the nn.Module
        target (str): the attribute you want to delete
    """
    *root, name = target.rsplit(".", 1)
    if root:
        root = root[0].split(".")
        delattr_recursively(getattr(module, root[0]), ".".join(root[1:] + [name]))
    else:
        delattr(module, target)


def fuse_linear_layers(
    model: nn.Module,
    modules: Dict[str, nn.Linear],
    targets: List[str],
    fused_layer_name: str,
) -> None:
    """Fuse two linear layers and append them to the nn Module.

    Args:
        model (nn.Module): Origianl nn.Module object
        modules (Dict[nn.Linear]): a dictiorany of node name: linear layer
        targets (List[str]): list of layer node names
        fused_layer_name (str): fused layer name

    Raises:
        ValueError: All linear layers must be of type nn.Linear and must have the same input dimension

    """
    # Get the attributes
    layers = [modules[name] for name in targets]

    in_features = list({layer.in_features for layer in layers})

    # ensure both linear layers have the same input dimensions and are not already fused
    if not all(isinstance(layer, nn.Linear) for layer in layers):
        raise ValueError("All linear layers must be of type nn.Linear")
    if len(in_features) != 1:
        raise ValueError(
            f"All linear layers must have the same input dimensions. Instead found: {in_features}"
        )

    # Create the new fused linear layer
    new_out_features = sum([layer.out_features for layer in layers])
    has_bias = any(layer.bias is not None for layer in layers)
    fused_layer = nn.Linear(in_features[0], new_out_features, bias=has_bias)

    # Concatenate the weights and biases
    with torch.no_grad():
        start, stop = 0, 0
        for layer in layers:
            stop += layer.out_features
            fused_layer.weight[start:stop, :] = layer.weight

            if has_bias:
                if layer.bias is not None:
                    fused_layer.bias[start:stop] = layer.bias
                else:
                    fused_layer.bias[start:stop] = torch.zeros_like(
                        fused_layer.bias[start:stop]
                    )
            start = stop

    # Replace the two layers in the original model with the new fused layer
    setattr(model, fused_layer_name, fused_layer)
    for layer_name in targets:
        delattr_recursively(model, layer_name)


def horizontal_fusion_linear(model: torch.nn.Module) -> torch.nn.Module:
    """Fuze horizontally two or more linear layers that share the same origin. This will increase NPU hw utilization.

    Args:
        model (torch.nn.Module): The original nn.Module

    Returns:
        torch.nn.Module: optimize nn.Module where parallel linear operations has been fused into a single bigger one
    """
    fx_model = fx.symbolic_trace(model)
    modules = dict(fx_model.named_modules())
    # new_graph = copy.deepcopy(fx_model.graph)

    def node_condition(node: Any) -> bool:
        """Return true if the node is a module and is nn.Linear.

        Args:
            node (Any): A torch fx node

        Returns:
            bool: return condition
        """
        return node.op == "call_module" and isinstance(modules[node.target], nn.Linear)

    # First, find all node with a linear layer
    linear_nodes = [node for node in fx_model.graph.nodes if node_condition(node)]

    # Group the linear layers by input node
    linear_nodes_parents: Dict[str, List[Any]] = {}
    for node in linear_nodes:
        linear_nodes_parents.setdefault(node.args[0], []).append(node)

    # Get the ones with size > 1
    fused_modules = [
        (source, modules)
        for source, modules in linear_nodes_parents.items()
        if len(modules) > 1
    ]

    for source, layers in fused_modules:
        fused_layer_name = "fused_" + "_".join(node.target for node in layers)
        fused_layer_name = fused_layer_name.replace(".", "_")
        fuse_linear_layers(
            fx_model, modules, [layer.target for layer in layers], fused_layer_name
        )
        with fx_model.graph.inserting_after(source):
            fused_node = fx_model.graph.call_module(fused_layer_name, (source,))

        with fx_model.graph.inserting_after(fused_node):

            start, stop = 0, 0
            for layer in layers:
                stop += modules[layer.target].out_features

                layer_slice = fx_model.graph.call_function(
                    operator.getitem,
                    args=(
                        fused_node,
                        (
                            Ellipsis,
                            slice(start, stop, None),
                        ),
                    ),
                    kwargs={},
                )
                layer.replace_all_uses_with(layer_slice)
                fx_model.graph.erase_node(layer)
                start = stop

    fx_model.graph.lint()
    fx_model.recompile()

    return fx_model
